defmodule Chronoscope.NTS.Behaviour do
  @callback(key_establishment(host :: String.t(), port :: integer()) :: {:ok, Map.t()}, {:error, any()})
end

defmodule Chronoscope.NTS do
  @behaviour Chronoscope.NTS.Behaviour

  require Logger

  alias Chronoscope.NTS

  def healthy?() do
    true
  end

  def list() do
    call_all_clients(:list)
  end

  def clear() do
    :terminate
    |> call_all_clients()
    |> Enum.map(&wait_for_termination/1)
  end

  defp wait_for_termination(pid) do
    ref = Process.monitor(pid)

    receive do
      {:DOWN, ^ref, _, _, _} -> {:ok, pid}
    after
      1000 -> {:error, pid}
    end
  end

  defp call_all_clients(message) do
    NTS.DynamicSupervisor
    |> DynamicSupervisor.which_children()
    |> Enum.map(fn {_, pid, _, _} -> GenServer.call(pid, message) end)
  end

  def remove(host, port) do
    name = "#{host}:#{port}"

    case Registry.lookup(NTS.Registry, name) do
      [{pid, _}] ->
        pid
        |> GenServer.call(:terminate)
        |> wait_for_termination()

      [] ->
        {:error, :notfound}
    end
  end

  @impl true
  def key_establishment(host, port) do
    GenServer.call(client_pid(host, port), :key_establishment)
  end

  defp client_pid(host, port) do
    name = "#{host}:#{port}"

    case Registry.lookup(NTS.Registry, name) do
      [{pid, _}] ->
        pid

      [] ->
        NTS.DynamicSupervisor
        |> DynamicSupervisor.start_child({NTS.Client, host: host, port: port, name: {:via, Registry, {NTS.Registry, name}}})
        |> then(fn {:ok, pid} -> pid end)
    end
  end
end
